from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class TalentOrb:
    def __init__(self, id: int, value: int):
        self.id = id
        self.value = value

    @staticmethod
    def init() -> TalentOrb:
        return TalentOrb(
            0,
            0,
        )

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> TalentOrb:
        id = stream.read_short()
        if gv < 110400:
            value = stream.read_byte()
        else:
            value = stream.read_short()
        return TalentOrb(id, value)

    def write(self, stream: core.Data, gv: core.GameVersion):
        stream.write_short(self.id)
        if gv < 110400:
            stream.write_byte(self.value)
        else:
            stream.write_short(self.value)

    def serialize(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "value": self.value,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> TalentOrb:
        return TalentOrb(data.get("id", 0), data.get("value", 0))

    def __repr__(self):
        return f"Orb({self.id}, {self.value})"

    def __str__(self):
        return self.__repr__()


class TalentOrbs:
    def __init__(self, orbs: dict[int, TalentOrb]):
        self.orbs = orbs

    @staticmethod
    def init() -> TalentOrbs:
        return TalentOrbs({})

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> TalentOrbs:
        length = stream.read_short()
        orbs: dict[int, TalentOrb] = {}
        for _ in range(length):
            orb = TalentOrb.read(stream, gv)
            orbs[orb.id] = orb
        return TalentOrbs(orbs)

    def write(self, stream: core.Data, gv: core.GameVersion):
        stream.write_short(len(self.orbs))
        for orb in self.orbs.values():
            orb.write(stream, gv)

    def serialize(self) -> list[dict[str, Any]]:
        return [orb.serialize() for orb in self.orbs.values()]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> TalentOrbs:
        return TalentOrbs(
            {orb.get("id", 0): TalentOrb.deserialize(orb) for orb in data}
        )

    def __repr__(self):
        return f"TalentOrbs({self.orbs})"

    def __str__(self):
        return self.__repr__()

    def set_orb(self, id: int, value: int):
        self.orbs[id] = TalentOrb(id, value)


class RawOrbInfo:
    def __init__(
        self,
        orb_id: int,
        rank_id: int,
        effect_id: int,
        value: list[int],
        target_id: int | None,
    ):
        self.orb_id = orb_id
        self.rank_id = rank_id
        self.effect_id = effect_id
        self.value = value
        self.target_id = target_id


class OrbInfo:
    def __init__(
        self,
        raw_orb_info: RawOrbInfo,
        rank: str,
        target: str | None,
        effect: str,
    ):
        self.raw_orb_info = raw_orb_info
        self.rank = rank
        self.target = target
        self.effect = effect

    def __str__(self) -> str:
        """Get the string representation of the OrbInfo

        Returns:
            str: The string representation of the OrbInfo
        """
        target_color = color_from_enemy_type(self.raw_orb_info.target_id)
        rank_color = color_from_grade(self.raw_orb_info.rank_id)
        effect_color = color_from_effect(self.raw_orb_info.effect_id)
        effect_text = self.effect.replace("%@", "{}")
        effect_text = f"<{effect_color}>{effect_text}</>"
        target = self.target
        effect = effect_text.format(
            f"<{rank_color}>{self.rank}</>",
            f"<{target_color}>{target}</>" if target else "",
        )
        return f"{effect}"

    def to_colortext(self) -> str:
        """Get the string representation of the OrbInfo with color

        Returns:
            str: The string representation of the OrbInfo with color
        """
        return str(self)

    @staticmethod
    def create_unknown(orb_id: int) -> OrbInfo:
        """Create an unknown OrbInfo

        Args:
            orb_id (int): The id of the orb

        Returns:
            OrbInfo: The unknown OrbInfo
        """
        return OrbInfo(
            RawOrbInfo(orb_id, 0, 0, [], 0),
            "???",
            "",
            "%@:%@",
        )


class OrbInfoList:
    equipment_data_file_name = "DataLocal/equipmentlist.json"
    grade_list_file_name = "DataLocal/equipmentgrade.csv"
    attribute_list_file_name = "resLocal/attribute_explonation.tsv"
    effect_list_file_name = "resLocal/equipment_explonation.tsv"

    def __init__(self, orb_info_list: list[OrbInfo]):
        """Initialize the OrbInfoList class

        Args:
            orb_info_list (list[OrbInfo]): The list of OrbInfo
        """
        self.orb_info_list = orb_info_list

    @staticmethod
    def create(save_file: core.SaveFile) -> OrbInfoList | None:
        """Create an OrbInfoList

        Args:
            save_file (core.SaveFile): The save file

        Returns:
            OrbInfoList | None: The OrbInfoList
        """
        gdg = core.core_data.get_game_data_getter(save_file)
        json_data_file = gdg.download_from_path(OrbInfoList.equipment_data_file_name)
        grade_list_file = gdg.download_from_path(OrbInfoList.grade_list_file_name)
        attribute_list_file = gdg.download_from_path(
            OrbInfoList.attribute_list_file_name
        )
        equipment_list_file = gdg.download_from_path(OrbInfoList.effect_list_file_name)
        if (
            json_data_file is None
            or grade_list_file is None
            or attribute_list_file is None
            or equipment_list_file is None
        ):
            return None
        raw_orbs = OrbInfoList.parse_json_data(json_data_file)
        if raw_orbs is None:
            return None
        orbs = OrbInfoList.load_names(
            raw_orbs, grade_list_file, attribute_list_file, equipment_list_file
        )
        return OrbInfoList(orbs)

    @staticmethod
    def parse_json_data(json_data: core.Data) -> list[RawOrbInfo] | None:
        """Parse the json data of the equipment

        Args:
            json_data (core.Data): The json data

        Returns:
            list[RawOrbInfo]: The list of RawOrbInfo
        """
        try:
            data: dict[str, Any] = core.JsonFile.from_data(json_data).to_object()
        except core.JSONDecodeError:
            return None
        orb_info_list: list[RawOrbInfo] = []
        for id, orb in enumerate(data["ID"]):
            grade_id = orb["gradeID"]
            content = orb["content"]
            value = orb["value"]
            attribute = orb.get("attribute")
            orb_info_list.append(RawOrbInfo(id, grade_id, content, value, attribute))
        return orb_info_list

    @staticmethod
    def load_names(
        raw_orb_info: list[RawOrbInfo],
        grade_data: core.Data,
        attribute_data: core.Data,
        effect_data: core.Data,
    ) -> list[OrbInfo]:
        """Load the names of the equipment

        Args:
            raw_orb_info (list[RawOrbInfo]): The list of RawOrbInfo
            grade_data (core.Data): Raw data of the grade list
            attribute_data (core.Data): Raw data of the attribute list
            effect_data (core.Data): Raw data of the effect list

        Returns:
            list[OrbInfo]: The list of OrbInfo
        """
        grade_csv = core.CSV(grade_data)
        attribute_tsv = core.CSV(attribute_data, "\t")
        effect_csv = core.CSV(effect_data, "\t")
        orb_info_list: list[OrbInfo] = []
        for orb in raw_orb_info:
            grade = grade_csv[orb.rank_id][3].to_str()
            effect = effect_csv[orb.effect_id][0].to_str()

            if orb.target_id is not None:
                attribute = attribute_tsv[orb.target_id][0].to_str()
            else:
                attribute = None

            orb_info_list.append(OrbInfo(orb, grade, attribute, effect))
        return orb_info_list

    def get_orb_info(self, orb_id: int) -> OrbInfo | None:
        """Get the OrbInfo from the id

        Args:
            orb_id (int): The id of the orb

        Returns:
            OrbInfo | None: The OrbInfo
        """
        try:
            return self.orb_info_list[orb_id]
        except IndexError:
            return None

    def get_orb_from_components(
        self,
        grade: str,
        attribute: str | None,
        effect: str,
    ) -> OrbInfo | None:
        """Get the OrbInfo from the components

        Args:
            grade (str): The grade of the orb
            attribute (str | None): The attribute of the orb. None if applies to all attributes
            effect (str): The effect of the orb

        Returns:
            OrbInfo | None: The OrbInfo
        """
        for orb in self.orb_info_list:
            if orb.rank == grade and orb.target == attribute and orb.effect == effect:
                return orb
        return None

    def does_match_orb_str(self, str_1: str | None, str_2: str | None) -> bool:
        if str_2 == "*":
            return True

        if str_1 is None:
            return str_2 is None
        if str_2 is None:
            return False

        return str_1.lower() == str_2.lower()

    def get_orbs_from_component_fuzzy(
        self,
        grade: str,
        attribute: str | None,
        effect: str,
    ) -> list[OrbInfo]:
        """Get the OrbInfo from the components matching the first word of the effect and lowercased

        Args:
            grade (str): The grade of the orb
            attribute (str | None): The attribute of the orb. None if all
            effect (str): The effect of the orb

        Returns:
            list[OrbInfo]: The list of OrbInfo
        """
        orbs: list[OrbInfo] = []
        for orb in self.orb_info_list:
            if (
                (orb.rank.lower() == grade.lower() or grade == "*")
                and (self.does_match_orb_str(orb.target, attribute))
                and (orb.effect == effect or effect == "*")
            ):
                orbs.append(orb)
        return orbs

    def get_all_grades(self) -> list[str]:
        """Get all the grades

        Returns:
            list[str]: The list of grades
        """

        data = list(
            set([(orb.rank, orb.raw_orb_info.rank_id) for orb in self.orb_info_list])
        )

        data.sort(key=lambda id: id[1])

        return [orb[0] for orb in data]

    def get_all_attributes(self) -> list[str | None]:
        """Get all the attributes

        Returns:
            list[str]: The list of attributes
        """

        data = list(
            set(
                [
                    (orb.target, orb.raw_orb_info.target_id)
                    for orb in self.orb_info_list
                    if orb.target is not None and orb.raw_orb_info.target_id is not None
                ]
            )
        )

        data.sort(key=lambda id: id[1])

        return [orb[0] for orb in data]

    def get_all_effects(self) -> list[str]:
        """Get all the effects

        Returns:
            list[str]: The list of effects
        """

        data = list(
            set(
                [(orb.effect, orb.raw_orb_info.effect_id) for orb in self.orb_info_list]
            )
        )

        data.sort(key=lambda id: id[1])

        return [orb[0] for orb in data]


class SaveOrb:
    """Represents a saved orb in the save file"""

    def __init__(self, orb: OrbInfo, count: int):
        """Initialize the SaveOrb class

        Args:
            orb (OrbInfo): The OrbInfo
            count (int): The amount of the orb
        """
        self.count = count
        self.orb = orb


def color_from_enemy_type(target_id: int | None) -> str:
    if target_id is None:
        return color.ColorHex.WHITE
    if target_id == 0:
        return color.ColorHex.RED
    elif target_id == 1:
        return color.ColorHex.GREEN
    elif target_id == 2:
        return color.ColorHex.DARK_GREY
    elif target_id == 3:
        return color.ColorHex.LIGHT_GREY
    elif target_id == 4:
        return color.ColorHex.YELLOW
    elif target_id == 5:
        return color.ColorHex.BLUE
    elif target_id == 6:
        return color.ColorHex.MAGENTA
    elif target_id == 7:
        return color.ColorHex.DARK_GREEN
    elif target_id == 8:
        return color.ColorHex.WHITE
    elif target_id == 9:
        return color.ColorHex.DARK_MAGENTA
    elif target_id == 10:
        return color.ColorHex.ORANGE
    elif target_id == 11:
        return color.ColorHex.CYAN
    return color.ColorHex.BLACK


def color_from_grade(grade_id: int) -> str:
    if grade_id == 0:
        return color.ColorHex.RED
    elif grade_id == 1:
        return color.ColorHex.ORANGE
    elif grade_id == 2:
        return color.ColorHex.YELLOW
    elif grade_id == 3:
        return color.ColorHex.GREEN
    elif grade_id == 4:
        return color.ColorHex.BLUE
    return color.ColorHex.BLACK


def color_from_effect(effect_id: int):
    if effect_id == 0:
        return color.ColorHex.RED
    elif effect_id == 1:
        return color.ColorHex.GREEN
    elif effect_id == 2:
        return color.ColorHex.DARK_GREY
    elif effect_id == 3:
        return color.ColorHex.LIGHT_GREY
    elif effect_id == 4:
        return color.ColorHex.YELLOW
    elif effect_id == 5:
        return color.ColorHex.BLUE
    elif effect_id == 6:
        return color.ColorHex.MAGENTA
    elif effect_id == 7:
        return color.ColorHex.DARK_GREEN
    elif effect_id == 8:
        return color.ColorHex.WHITE
    elif effect_id == 9:
        return color.ColorHex.DARK_MAGENTA
    elif effect_id == 10:
        return color.ColorHex.ORANGE

    return color.ColorHex.BLACK


class SaveOrbs:
    def __init__(
        self,
        orbs: dict[int, SaveOrb],
        orb_info_list: OrbInfoList,
    ):
        """Initialize the SaveOrbs class

        Args:
            orbs (dict[int, SaveOrb]): The orbs
            orb_info_list (OrbInfoList): The orb info list
        """
        self.orbs = orbs
        self.orb_info_list = orb_info_list

    @staticmethod
    def from_save_file(save_file: core.SaveFile) -> SaveOrbs | None:
        """Create a SaveOrbs from the save stats

        Args:
            save_file (core.SaveFile): The save file

        Returns:
            SaveOrbs | None: The SaveOrbs
        """
        orb_info_list = OrbInfoList.create(save_file)
        if orb_info_list is None:
            return None
        orbs: dict[int, SaveOrb] = {}
        for orb_id, orb in save_file.talent_orbs.orbs.items():
            try:
                orb_info = orb_info_list.orb_info_list[int(orb_id)]
            except IndexError:
                orb_info = OrbInfo.create_unknown(int(orb_id))
            orbs[int(orb_id)] = SaveOrb(orb_info, orb.value)

        return SaveOrbs(orbs, orb_info_list)

    def print(self):
        """Print the orbs as a formatted list"""
        self.sort_orbs()
        total_orbs = sum([orb.count for orb in self.orbs.values()])
        color.ColoredText.localize("total_current_orbs", total_orbs=total_orbs)
        color.ColoredText.localize(
            "total_current_orb_types", total_types=len(self.orbs)
        )
        color.ColoredText.localize("current_orbs")
        for orb in self.orbs.values():
            color.ColoredText(f"<@q>{orb.count}</> {orb.orb.to_colortext()}")

    def sort_orbs(self):
        """Sort the orbs by attribute, effect, grade and id in that order with attribute being the most important"""
        orbs = list(self.orbs.values())
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.orb_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.rank_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.effect_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.target_id or -1)

    def localize_attribute(self, attribute: str | None) -> str | None:
        if attribute is not None:
            return attribute

    def edit(self):
        """Edit the orbs"""
        # this code sucks quit a lot, but it works and i can't be bothered making it better atm
        self.print()
        all_grades = self.orb_info_list.get_all_grades()
        all_grades = [grade for grade in all_grades]
        all_grades.sort()
        all_attributes = self.orb_info_list.get_all_attributes()
        all_attributes = [
            self.localize_attribute(attribute) or ""
            for attribute in all_attributes
            if attribute
        ]
        all_attributes.sort()
        all_effects = self.orb_info_list.get_all_effects()
        all_effects.sort()
        all_effects_str = [
            effect.lower().replace("%@", "").replace(":", "").strip() + f" ({i})"
            for (i, effect) in enumerate(all_effects)
        ]
        all_effect_ids = [i for i in range(len(all_effects))]

        all_grades_str = "".join(
            f"<{color_from_grade(self.orb_info_list.get_all_grades().index(grade))}>{grade}</>,"
            for grade in all_grades
        )

        all_attributes_str = "".join(
            f"<{color_from_enemy_type(self.orb_info_list.get_all_attributes().index(attribute))}>{attribute}</>,"
            for attribute in all_attributes
        )

        all_effects_str = "".join(
            f"<{color_from_effect(self.orb_info_list.get_all_effects().index(effect))}>{effect_str}</>,"
            for effect_str, effect in zip(all_effects_str, all_effects)
        )

        color.ColoredText.localize(
            "edit_orbs_help",
            escape=False,
            all_grades_str=all_grades_str,
            all_attributes_str=all_attributes_str,
            all_effects_str=all_effects_str,
        )

        orb_input_selection = (
            color.ColoredInput()
            .localize("orb_select")
            .lower()
            .replace("angle", "angel")
            .split(",")
        )
        if orb_input_selection == [core.core_data.local_manager.get_key("quit_key")]:
            return

        orb_selection: list[OrbInfo] = []

        for orb_input in orb_input_selection:
            grade = None
            attribute = None
            effect = None
            orb_input = orb_input.strip()
            parts = orb_input.split(" ")
            parts = [part.lower() for part in parts if part != ""]
            if len(parts) == 0:
                continue
            if parts[0] == "*":
                orb_selection = self.orb_info_list.orb_info_list
                break
            for available_grade in all_grades:
                if available_grade.lower() in parts:
                    grade = available_grade
                    break
            for available_attribute in all_attributes:
                if available_attribute.lower() in parts:
                    attribute = available_attribute
                    break
            for available_effect in all_effect_ids:
                if str(available_effect) in parts:
                    effect = all_effects[available_effect]
                    break
            if grade is None:
                grade = "*"
            if attribute is None:
                attribute = "*"
            if effect is None:
                effect = "*"
            orbs = self.orb_info_list.get_orbs_from_component_fuzzy(
                grade, attribute, effect
            )
            orb_selection.extend(orbs)

        orb_selection = list(set(orb_selection))
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.orb_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.rank_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.effect_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.target_id or -1)

        color.ColoredText.localize("selected_orbs")

        for orb in orb_selection:
            color.ColoredText(orb.to_colortext())

        max_orbs = core.core_data.max_value_manager.get("talent_orbs")

        if len(orb_selection) == 0:
            return
        if len(orb_selection) == 1:
            individual = True
        else:
            individual = dialog_creator.ChoiceInput.from_reduced(
                ["individual", "edit_all_at_once"],
                dialog="edit_orbs_individually",
                single_choice=True,
            ).single_choice()
            if individual is None:
                return
            individual = True if individual == 1 else False
        if individual:
            for orb in orb_selection:
                orb_id = orb.raw_orb_info.orb_id
                try:
                    orb_count = self.orbs[orb_id].count
                except KeyError:
                    orb_count = 0

                orb_count = dialog_creator.SingleEditor(
                    orb.to_colortext(), orb_count, max_orbs
                ).edit(escape_text=False)

                self.orbs[orb_id] = SaveOrb(orb, orb_count)

        else:
            int_input = dialog_creator.IntInput(max_orbs)
            orb_count = int_input.get_input_locale_while(
                "edit_orbs_all", {"max": max_orbs}, escape=False
            )
            if orb_count is None:
                return
            orb_count = int_input.clamp_value(orb_count)
            for orb in orb_selection:
                orb_id = orb.raw_orb_info.orb_id
                self.orbs[orb_id] = SaveOrb(orb, orb_count)

        self.print()

    def save(self, save_file: core.SaveFile):
        """Save the orbs to the save_stats

        Args:
            save_file (core.SaveFile): The save_stats to save the orbs to
        """
        for orb_id, orb in self.orbs.items():
            save_file.talent_orbs.orbs[orb_id] = core.TalentOrb(orb_id, orb.count)

    @staticmethod
    def edit_talent_orbs(save_file: core.SaveFile):
        """Edit the talent orbs

        Args:
            save_file (core.SaveFile): The save_stats to edit the orbs of

        """
        save_orbs = SaveOrbs.from_save_file(save_file)
        if save_orbs is None:
            color.ColoredText.localize("failed_to_load_orbs")
            return None
        save_orbs.edit()
        save_orbs.save(save_file)



from __future__ import annotations, division
from bcsfe import core

from typing import Any

from bcsfe.cli import dialog_creator, color


class SpecialSkill:
    def __init__(self, upg: core.Upgrade):
        self.upgrade = upg
        self.seen = 0
        self.max_upgrade_level = core.Upgrade(0, 0)

    @staticmethod
    def init() -> SpecialSkill:
        return SpecialSkill(core.Upgrade(0, 0))

    @staticmethod
    def read_upgrade(stream: core.Data) -> SpecialSkill:
        up = core.Upgrade.read(stream)
        return SpecialSkill(up)

    def write_upgrade(self, stream: core.Data):
        self.upgrade.write(stream)

    def read_seen(self, stream: core.Data):
        self.seen = stream.read_int()

    def write_seen(self, stream: core.Data):
        stream.write_int(self.seen)

    def read_max_upgrade_level(self, stream: core.Data):
        level = core.Upgrade.read(stream)
        self.max_upgrade_level = level

    def write_max_upgrade_level(self, stream: core.Data):
        self.max_upgrade_level.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "upgrade": self.upgrade.serialize(),
            "seen": self.seen,
            "max_upgrade_level": self.max_upgrade_level.serialize(),
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> SpecialSkill:
        skill = SpecialSkill(core.Upgrade.deserialize(data.get("upgrade", {})))
        skill.seen = data.get("seen", 0)
        skill.max_upgrade_level = core.Upgrade.deserialize(
            data.get("max_upgrade_level", {})
        )
        return skill

    def __repr__(self) -> str:
        return f"Skill(upgrade={self.upgrade}, seen={self.seen}, max_upgrade_level={self.max_upgrade_level})"

    def __str__(self) -> str:
        return self.__repr__()

    def set_upgrade(
        self,
        upgrade: core.Upgrade,
        only_plus: bool = False,
        max_base: int | None = None,
        max_plus: int | None = None,
    ):
        if max_base is not None:
            upgrade.base = min(upgrade.base, max_base)
        if max_plus is not None:
            upgrade.plus = min(upgrade.plus, max_plus)

        base = upgrade.base
        plus = upgrade.plus

        if base != -1 and not only_plus:
            self.upgrade.base = upgrade.get_random_base(max_base)
        if plus != -1:
            self.upgrade.plus = upgrade.get_random_plus(max_plus)


class SpecialSkills:
    def __init__(self, skills: list[SpecialSkill]):
        self.skills = skills

    def get_upgrade(self, valid_skill_id: int) -> SpecialSkill:
        if valid_skill_id >= 1:
            valid_skill_id += 1

        return self.skills[valid_skill_id]

    def set_upgrade(
        self,
        valid_skill_id: int,
        upgrade: core.Upgrade,
        max_base: int | None = None,
        max_plus: int | None = None,
    ):
        u = upgrade.copy()
        valid_skills = self.get_valid_skills()
        valid_skills[valid_skill_id].set_upgrade(
            u, max_base=max_base, max_plus=max_plus
        )

        if (
            valid_skill_id == 0
        ):  # if it is a cat cannon power upgrade, mirror the upgrade to the hidden cat cannon power special skill
            self.skills[1].set_upgrade(u, max_base=max_base, max_plus=max_plus)

    @staticmethod
    def init() -> SpecialSkills:
        skills = [SpecialSkill.init() for _ in range(11)]
        return SpecialSkills(skills)

    def get_valid_skills(self) -> list[SpecialSkill]:
        new_skills: list[SpecialSkill] = []
        for i, skill in enumerate(self.skills):
            if i == 1:
                continue
            new_skills.append(skill)

        return new_skills

    @staticmethod
    def read_upgrades(stream: core.Data) -> SpecialSkills:
        total_skills = 11

        skills: list[SpecialSkill] = []
        for _ in range(total_skills):
            skills.append(SpecialSkill.read_upgrade(stream))

        return SpecialSkills(skills)

    def write_upgrades(self, stream: core.Data):
        for skill in self.skills:
            skill.write_upgrade(stream)

    def read_gatya_seen(self, stream: core.Data):
        for skill in self.get_valid_skills():
            skill.read_seen(stream)

    def write_gatya_seen(self, stream: core.Data):
        for skill in self.get_valid_skills():
            skill.write_seen(stream)

    def read_max_upgrade_levels(self, stream: core.Data):
        for skill in self.skills:
            skill.read_max_upgrade_level(stream)

    def write_max_upgrade_levels(self, stream: core.Data):
        for skill in self.skills:
            skill.write_max_upgrade_level(stream)

    def serialize(self) -> list[dict[str, Any]]:
        return [skill.serialize() for skill in self.skills]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> SpecialSkills:
        skills = SpecialSkills([])
        for skill in data:
            skills.skills.append(SpecialSkill.deserialize(skill))

        return skills

    def __repr__(self) -> str:
        return f"Skills(skills={self.skills})"

    def __str__(self) -> str:
        return f"Skills(skills={self.skills})"

    def edit(self, save_file: core.SaveFile):
        names_o = core.core_data.get_gatya_item_names(save_file)
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(2)
        if items is None:
            return
        names: list[str] = []
        for item in items:
            name = names_o.get_name(item.id)
            if name is None:
                return
            names.append(name)
        ids, _ = dialog_creator.ChoiceInput.from_reduced(
            names, [], {}, "special_skills_dialog"
        ).multiple_choice()
        if not ids:
            return
        skills = self.get_valid_skills()
        if len(ids) == 1:
            option_id = 0
        else:
            options: list[str] = [
                "upgrade_individual_skill",
                "upgrade_all_skills",
            ]
            option_id = dialog_creator.ChoiceInput(
                options, options, [], {}, "upgrade_skills_select_mod", True
            ).single_choice()
            if option_id is None:
                return
            option_id -= 1

        ability_data = core.core_data.get_ability_data(save_file)
        if ability_data.ability_data is None:
            return
        success = False
        if option_id == 0:
            for id in ids:
                color.ColoredText.localize(
                    "selected_skill_upgrades",
                    name=names[id],
                    base_level=skills[id].upgrade.base + 1,
                    plus_level=skills[id].upgrade.plus,
                )
                ability = ability_data.get_ability_data_item(id)
                if ability is None:
                    continue
                upgrade, should_exit = core.Upgrade.get_user_upgrade(
                    ability.max_base_level - 1, ability.max_plus_level
                )
                if should_exit:
                    return
                if upgrade is not None:
                    self.set_upgrade(id, upgrade)
                    color.ColoredText.localize(
                        "selected_skill_upgraded",
                        name=names[id],
                        base_level=skills[id].upgrade.base + 1,
                        plus_level=skills[id].upgrade.plus,
                    )
                    success = True

        elif option_id == 1:
            max_base_level = max(
                [ability.max_base_level for ability in ability_data.ability_data]
            )
            max_plus_level = max(
                [ability.max_plus_level for ability in ability_data.ability_data]
            )
            upgrade, should_exit = core.Upgrade.get_user_upgrade(
                max_base_level - 1, max_plus_level
            )
            if should_exit or upgrade is None:
                return
            disable_maxes = core.core_data.config.get_bool(core.ConfigKey.DISABLE_MAXES)
            for id in ids:
                max_base_level = ability_data.ability_data[id].max_base_level - 1
                max_plus_level = ability_data.ability_data[id].max_plus_level
                if disable_maxes:
                    max_base_level = None
                    max_plus_level = None

                self.set_upgrade(
                    id,
                    upgrade.copy(),
                    max_base=max_base_level,
                    max_plus=max_plus_level,
                )

                color.ColoredText.localize(
                    "selected_skill_upgraded",
                    name=names[id],
                    base_level=skills[id].upgrade.base + 1,
                    plus_level=skills[id].upgrade.plus,
                )
            success = True

        if success:
            color.ColoredText.localize("skills_edited")

    def get_from_id(self, id: int, only_valid: bool = True) -> SpecialSkill | None:
        if only_valid:
            skills = self.get_valid_skills()
        else:
            skills = self.skills
        if id >= len(skills) or id < 0:
            return None
        return skills[id]


class AbilityDataItem:
    def __init__(
        self,
        index: int,
        sell_price: int,
        gatya_rarity: int,
        max_base_level: int,
        max_plus_level: int,
        chapter_1_to_2_max_level: int,
    ):
        self.index = index
        self.sell_price = sell_price
        self.gatya_rarity = gatya_rarity
        self.max_base_level = max_base_level
        self.max_plus_level = max_plus_level
        self.chapter_1_to_2_max_level = chapter_1_to_2_max_level


class AbilityData:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.ability_data = self.get_ability_data()

    def get_ability_data(self) -> list[AbilityDataItem] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "AbilityData.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        ability_data: list[AbilityDataItem] = []
        for i, row in enumerate(csv):
            ability_data.append(
                AbilityDataItem(
                    i,
                    row[0].to_int(),
                    row[1].to_int(),
                    row[2].to_int(),
                    row[3].to_int(),
                    row[4].to_int(),
                )
            )
        return ability_data

    def get_ability_data_item(self, item_id: int) -> AbilityDataItem | None:
        if self.ability_data is None:
            return None
        return self.ability_data[item_id]



from __future__ import annotations
from dataclasses import dataclass

from bcsfe import core


@dataclass
class Drop:
    stage_id: int
    save_id: int
    chara_id: int


class CharaDrop:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.drops = self.get_drops()

    def get_drops(self) -> list[Drop] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "drop_chara.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        drops: list[Drop] = []
        for line in csv.lines[1:]:
            drops.append(
                Drop(
                    stage_id=line[0].to_int(),
                    save_id=line[1].to_int(),
                    chara_id=line[2].to_int(),
                )
            )

        return drops

    def get_drop(self, stage_id: int) -> Drop | None:
        if self.drops is None:
            return None
        for drop in self.drops:
            if drop.stage_id == stage_id:
                return drop

        return None

    def get_drops_from_chara_id(self, chara_id: int) -> list[Drop] | None:
        if self.drops is None:
            return None
        drops: list[Drop] = []
        for drop in self.drops:
            if drop.chara_id == chara_id:
                drops.append(drop)

        return drops

    def unlock_drops_from_cat_id(self, cat_id: int) -> None:
        drops = self.get_drops_from_chara_id(cat_id)
        if drops is None:
            return
        for drop in drops:
            try:
                self.save_file.unit_drops[drop.save_id] = 1
            except IndexError:
                pass

    def remove_drops_from_cat_id(self, cat_id: int) -> None:
        drops = self.get_drops_from_chara_id(cat_id)
        if drops is None:
            return
        for drop in drops:
            try:
                self.save_file.unit_drops[drop.save_id] = 0
            except IndexError:
                pass



from __future__ import annotations
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class FileDialog:
    def load_tk(self):
        try:
            import tkinter as tk
            from tkinter import filedialog

            self.tk = tk
            self.filedialog = filedialog
        except ImportError:
            self.tk = None
            self.filedialog = None

    def __init__(self):
        self.load_tk()
        if self.tk is not None:
            try:
                self.root = self.tk.Tk()
            except self.tk.TclError:
                self.tk = None
                self.filedialog = None
                return

            self.root.withdraw()
            self.root.wm_attributes("-topmost", 1)  # type: ignore

    def select_files_in_dir(
        self, path: core.Path, ignore_json: bool
    ) -> str | None:
        """Print current files in directory.

        Args:
            path (core.Path): Path to directory.
        """
        color.ColoredText.localize("current_files_dir", dir=path)
        path.generate_dirs()
        files = path.get_files()
        if not files:
            color.ColoredText.localize("no_files_dir")

        files.sort(key=lambda file: file.basename())

        # remove files with .json extension
        if ignore_json:
            files = [file for file in files if file.get_extension() != "json"]

        files_str_ls = [file.basename() for file in files]
        options = files_str_ls + [core.localize("other_dir"), core.localize("another_path")]

        choice = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="select_files_dir",
            single_choice=True,
            localize_options=False,
        ).single_choice()
        if choice is None:
            return

        choice -= 1
        if choice == len(files):
            path_input = color.ColoredInput().localize("enter_path_dir")
            path_obj = core.Path(path_input)
            if path_obj.is_relative():
                path_obj = path.add(path_obj)
            if not path_obj.exists():
                color.ColoredText.localize("path_not_exists", path=path_obj)
                return self.select_files_in_dir(path, ignore_json)
            return self.select_files_in_dir(path_obj, ignore_json)
        if choice == len(files) + 1:
            path_input = color.ColoredInput().localize("enter_path")
            return path_input or None
        return files[choice].to_str()

    def use_tk(self) -> bool:
        return (
            self.tk is not None
            and self.filedialog is not None
            and core.core_data.config.get_bool(core.ConfigKey.USE_FILE_DIALOG)
        )

    def get_file(
        self,
        title: str,
        initialdir: str,
        initialfile: str,
        filetypes: list[tuple[str, str]] | None = None,
        ignore_json: bool = False,
    ) -> str | None:
        if filetypes is None:
            filetypes = []
        title = core.core_data.local_manager.get_key(title)
        color.ColoredText.localize(title)
        if not self.use_tk():
            curr_path = core.Path(initialdir).add(initialfile)
            file = self.select_files_in_dir(curr_path.parent(), ignore_json)
            if file is None:
                return None
            path_obj = core.Path(file)
            if path_obj.exists():
                return file
            color.ColoredText.localize("path_not_exists", path=path_obj)
            return None

        return (
            self.filedialog.askopenfilename(  # type: ignore
                title=title,
                filetypes=filetypes,
                initialdir=initialdir,
                initialfile=initialfile,
            )
            or None
        )

    def save_file(
        self,
        title: str,
        initialdir: str,
        initialfile: str,
        filetypes: list[tuple[str, str]] | None = None,
    ) -> str | None:
        """Save file dialog

        Args:
            title (str): Title of dialog.
            filetypes (list[tuple[str, str]] | None, optional): File types. Defaults to None.
            initialdir (str, optional): Initial directory. Defaults to "".
            initialfile (str, optional): Initial file. Defaults to "".

        Returns:
            str | None: Path to file.
        """
        if filetypes is None:
            filetypes = []
        title = core.core_data.local_manager.get_key(title)
        color.ColoredText.localize(title)
        if not self.use_tk():
            def_path = core.Path(initialdir).add(initialfile).to_str()
            path = color.ColoredInput().localize(
                "enter_path_default", default=def_path
            )
            return path.strip().strip("'").strip('"') if path else def_path
        return (
            self.filedialog.asksaveasfilename(  # type: ignore
                title=title,
                filetypes=filetypes,
                initialdir=initialdir,
                initialfile=initialfile,
            )
            or None
        )



from __future__ import annotations
import traceback

from bcsfe import cli

from bcsfe import core

import bcsfe
import argparse


def main():
    parser = argparse.ArgumentParser("bcsfe")

    parser.add_argument(
        "--version", "-v", action="store_true", help="display the version and exit"
    )
    parser.add_argument(
        "--input-path", "-i", type=str, help="input path to save file to edit"
    )
    parser.add_argument(
        "--config-path",
        "-c",
        type=str,
        default=None,
        help="path to the config file. If unspecified defaults to Documents/bcsfe/config.yaml",
    )
    parser.add_argument(
        "--log-path",
        "-l",
        type=str,
        default=None,
        help="path to the log file. If unspecified defaults to Documents/bcsfe/bcsfe.log",
    )

    args = parser.parse_args()
    if args.version:
        print(bcsfe.__version__)
        exit()

    if args.config_path is not None:
        core.set_config_path(core.Path(args.config_path))

    if args.log_path is not None:
        core.set_log_path(core.Path(args.log_path))

    core.core_data.init_data()

    try:
        cli.main.Main().main(args.input_path)
    except KeyboardInterrupt:
        cli.main.Main.leave()
    except Exception as e:
        tb = traceback.format_exc()
        cli.color.ColoredText.localize(
            "error", error=e, version=bcsfe.__version__, traceback=tb
        )
        try:
            cli.main.Main.exit_editor()
        except Exception:
            pass
        except KeyboardInterrupt:
            pass


main()



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class Stage:
    def __init__(self, clear_times: int):
        self.clear_times = clear_times
        self.treasure = 0
        self.itf_timed_score = 0

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read_clear_times(stream: core.Data) -> Stage:
        return Stage(stream.read_int())

    def read_treasure(self, stream: core.Data):
        self.treasure = stream.read_int()

    def read_itf_timed_score(self, stream: core.Data):
        self.itf_timed_score = stream.read_int()

    def write_clear_times(self, stream: core.Data):
        stream.write_int(self.clear_times)

    def write_treasure(self, stream: core.Data):
        stream.write_int(self.treasure)

    def write_itf_timed_score(self, stream: core.Data):
        stream.write_int(self.itf_timed_score)

    def serialize(self) -> dict[str, Any]:
        return {
            "clear_times": self.clear_times,
            "treasure": self.treasure,
            "itf_timed_score": self.itf_timed_score,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Stage:
        stage = Stage(data.get("clear_times", 0))
        stage.treasure = data.get("treasure", 0)
        stage.itf_timed_score = data.get("itf_timed_score", 0)
        return stage

    def __repr__(self):
        return f"Stage({self.clear_times}, {self.treasure}, {self.itf_timed_score})"

    def __str__(self):
        return self.__repr__()

    def clear_stage(self, clear_amount: int = 1):
        self.clear_times = clear_amount

    def unclear_stage(self):
        self.clear_times = 0

    def is_cleared(self) -> bool:
        return self.clear_times > 0

    def set_treasure(self, treasure: int):
        self.treasure = treasure


class Chapter:
    def __init__(self, selected_stage: int):
        self.selected_stage = selected_stage
        self.progress = 0
        self.stages = [Stage.init() for _ in range(51)]
        self.time_until_treasure_chance = 0
        self.treasure_chance_duration = 0
        self.treasure_chance_value = 0
        self.treasure_chance_stage_id = 0
        self.treasure_festival_type = 0

    def clear_stage(
        self,
        index: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
    ):
        if overwrite_clear_progress:
            self.progress = index + 1
        else:
            self.progress = max(self.progress, index + 1)
        self.stages[index].clear_stage(clear_amount)

    def set_treasure(self, stage_id: int, treasure: int):
        self.stages[stage_id].set_treasure(treasure)

    def is_stage_clear(self, stage_id: int) -> bool:
        return self.stages[stage_id].is_cleared()

    @staticmethod
    def init() -> Chapter:
        return Chapter(0)

    def get_treasure_stages(self) -> list[Stage]:
        return self.stages[:49]

    def get_valid_treasure_stages(self) -> list[Stage]:
        return self.stages[:48]

    @staticmethod
    def read_selected_stage(stream: core.Data) -> Chapter:
        return Chapter(stream.read_int())

    def read_progress(self, stream: core.Data):
        self.progress = stream.read_int()

    def read_clear_times(self, stream: core.Data):
        total_stages = 51
        self.stages = [Stage.read_clear_times(stream) for _ in range(total_stages)]

    def read_treasure(self, stream: core.Data):
        for stage in self.get_treasure_stages():
            stage.read_treasure(stream)

    def read_time_until_treasure_chance(self, stream: core.Data):
        self.time_until_treasure_chance = stream.read_int()

    def read_treasure_chance_duration(self, stream: core.Data):
        self.treasure_chance_duration = stream.read_int()

    def read_treasure_chance_value(self, stream: core.Data):
        self.treasure_chance_value = stream.read_int()

    def read_treasure_chance_stage_id(self, stream: core.Data):
        self.treasure_chance_stage_id = stream.read_int()

    def read_treasure_festival_type(self, stream: core.Data):
        self.treasure_festival_type = stream.read_int()

    def read_itf_timed_scores(self, stream: core.Data):
        for stage in self.stages:
            stage.read_itf_timed_score(stream)

    def write_selected_stage(self, stream: core.Data):
        stream.write_int(self.selected_stage)

    def write_progress(self, stream: core.Data):
        stream.write_int(self.progress)

    def write_clear_times(self, stream: core.Data):
        for stage in self.stages:
            stage.write_clear_times(stream)

    def write_treasure(self, stream: core.Data):
        for stage in self.get_treasure_stages():
            stage.write_treasure(stream)

    def write_time_until_treasure_chance(self, stream: core.Data):
        stream.write_int(self.time_until_treasure_chance)

    def write_treasure_chance_duration(self, stream: core.Data):
        stream.write_int(self.treasure_chance_duration)

    def write_treasure_chance_value(self, stream: core.Data):
        stream.write_int(self.treasure_chance_value)

    def write_treasure_chance_stage_id(self, stream: core.Data):
        stream.write_int(self.treasure_chance_stage_id)

    def write_treasure_festival_type(self, stream: core.Data):
        stream.write_int(self.treasure_festival_type)

    def write_itf_timed_scores(self, stream: core.Data):
        for stage in self.stages:
            stage.write_itf_timed_score(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "selected_stage": self.selected_stage,
            "progress": self.progress,
            "stages": [stage.serialize() for stage in self.stages],
            "time_until_treasure_chance": self.time_until_treasure_chance,
            "treasure_chance_duration": self.treasure_chance_duration,
            "treasure_chance_value": self.treasure_chance_value,
            "treasure_chance_stage_id": self.treasure_chance_stage_id,
            "treasure_festival_type": self.treasure_festival_type,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Chapter:
        chapter = Chapter(data.get("selected_stage", 0))
        chapter.progress = data.get("progress", 0)
        chapter.stages = [Stage.deserialize(stage) for stage in data.get("stages", [])]
        chapter.time_until_treasure_chance = data.get("time_until_treasure_chance", 0)
        chapter.treasure_chance_duration = data.get("treasure_chance_duration", 0)
        chapter.treasure_chance_value = data.get("treasure_chance_value", 0)
        chapter.treasure_chance_stage_id = data.get("treasure_chance_stage_id", 0)
        chapter.treasure_festival_type = data.get("treasure_festival_type", 0)
        return chapter

    def __repr__(self):
        return f"Chapter({self.selected_stage}, {self.progress}, {self.stages}, {self.time_until_treasure_chance}, {self.treasure_chance_duration}, {self.treasure_chance_value}, {self.treasure_chance_stage_id}, {self.treasure_festival_type})"

    def __str__(self):
        return f"Chapter({self.selected_stage}, {self.progress}, {self.stages}, {self.time_until_treasure_chance}, {self.treasure_chance_duration}, {self.treasure_chance_value}, {self.treasure_chance_stage_id}, {self.treasure_festival_type})"

    def apply_progress(self, progress: int, clear_times: list[int] | None = None):
        if clear_times is None:
            clear_times = [1] * progress

        self.progress = progress
        for i in range(progress + 1, 48):
            self.stages[i].unclear_stage()

        for i in range(progress):
            self.stages[i].clear_stage(clear_times[i])

    def clear_chapter(self):
        self.apply_progress(48)


class StoryChapters:
    def __init__(self, chapters: list[Chapter]):
        self.chapters = chapters

    def get_real_chapters(self) -> list[Chapter]:
        new_chapters: list[Chapter] = []
        for i, chapter in enumerate(self.chapters):
            if i == 3:
                continue
            new_chapters.append(chapter)
        return new_chapters

    def clear_stage(
        self,
        map: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        chapters: list[Chapter] | None = None,
    ):
        if chapters is None:
            chapters = self.chapters
        chapters[map].clear_stage(stage, clear_amount, overwrite_clear_progress)

    def set_treasure(self, chapter: int, stage: int, treasure: int):
        self.chapters[chapter].set_treasure(stage, treasure)

    def is_stage_clear(self, chapter: int, stage: int) -> bool:
        return self.chapters[chapter].is_stage_clear(stage)

    @staticmethod
    def init() -> StoryChapters:
        chapters = [Chapter.init() for _ in range(10)]
        return StoryChapters(chapters)

    @staticmethod
    def read(stream: core.Data) -> StoryChapters:
        total_chapters = 10
        chapters_l = [
            Chapter.read_selected_stage(stream) for _ in range(total_chapters)
        ]
        chapters = StoryChapters(chapters_l)
        for chapter in chapters.chapters:
            chapter.read_progress(stream)
        for chapter in chapters.chapters:
            chapter.read_clear_times(stream)
        for chapter in chapters.chapters:
            chapter.read_treasure(stream)
        return chapters

    def read_treasure_festival(self, stream: core.Data):
        for chapter in self.chapters:
            chapter.read_time_until_treasure_chance(stream)
        for chapter in self.chapters:
            chapter.read_treasure_chance_duration(stream)
        for chapter in self.chapters:
            chapter.read_treasure_chance_value(stream)
        for chapter in self.chapters:
            chapter.read_treasure_chance_stage_id(stream)
        for chapter in self.chapters:
            chapter.read_treasure_festival_type(stream)

    def write(self, stream: core.Data):
        for chapter in self.chapters:
            chapter.write_selected_stage(stream)
        for chapter in self.chapters:
            chapter.write_progress(stream)
        for chapter in self.chapters:
            chapter.write_clear_times(stream)
        for chapter in self.chapters:
            chapter.write_treasure(stream)

    def write_treasure_festival(self, stream: core.Data):
        for chapter in self.chapters:
            chapter.write_time_until_treasure_chance(stream)
        for chapter in self.chapters:
            chapter.write_treasure_chance_duration(stream)
        for chapter in self.chapters:
            chapter.write_treasure_chance_value(stream)
        for chapter in self.chapters:
            chapter.write_treasure_chance_stage_id(stream)
        for chapter in self.chapters:
            chapter.write_treasure_festival_type(stream)

    def read_itf_timed_scores(self, stream: core.Data):
        # 0: eoc 1
        # 1: eoc 2
        # 2: eoc 3
        # 3: _
        # 4: itf 1
        # 5: itf 2
        # 6: itf 3
        # 7: cotc 1
        # 8: cotc 2
        # 9: cotc 3

        for i, chapter in enumerate(self.chapters):
            if i > 3 and i < 7:
                chapter.read_itf_timed_scores(stream)

    def write_itf_timed_scores(self, stream: core.Data):
        for i, chapter in enumerate(self.chapters):
            if i > 3 and i < 7:
                chapter.write_itf_timed_scores(stream)

    def serialize(self) -> list[dict[str, Any]]:
        chapters = [chapter.serialize() for chapter in self.chapters]
        return chapters

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> StoryChapters:
        chapters = StoryChapters([Chapter.deserialize(chapter) for chapter in data])
        return chapters

    def __repr__(self):
        return f"Chapters({self.chapters})"

    def __str__(self):
        return f"Chapters({self.chapters})"

    @staticmethod
    def clear_tutorial(save_file: core.SaveFile):
        save_file.tutorial_state = max(save_file.tutorial_state, 1)
        save_file.koreaSuperiorTreasureState = max(
            save_file.koreaSuperiorTreasureState, 2
        )
        save_file.ui6 = max(save_file.ui6, 1)
        new_length = len(save_file.new_dialogs_2)
        if new_length < 6:
            save_file.new_dialogs_2.extend([0] * (6 - new_length))

        save_file.new_dialogs_2[1] = max(save_file.new_dialogs_2[1], 2)
        save_file.new_dialogs_2[5] = max(save_file.new_dialogs_2[5], 2)
        if save_file.story.chapters[0].stages[0].clear_times == 0:
            save_file.story.clear_stage(0, 0)

    @staticmethod
    def get_chapter_names(
        save_file: core.SaveFile, chapter_ids: list[int] | None = None
    ) -> list[str] | None:
        if chapter_ids is None:
            chapter_ids = [0, 1, 2, 3, 4, 5, 6, 7, 8]

        chapter_names: list[str] = []
        localizable = core.core_data.get_localizable(save_file)
        eoc_name = localizable.get("everyplay_mapname_J")
        itf_name = localizable.get("everyplay_mapname_W")
        cotc_name = localizable.get("everyplay_mapname_P")
        if eoc_name is None or itf_name is None or cotc_name is None:
            return None

        for chapter_id in chapter_ids:
            if chapter_id < 3:
                chapter_names.append(eoc_name.replace("%d", str(chapter_id + 1)))
            elif chapter_id < 6:
                chapter_names.append(itf_name.replace("%d", str(chapter_id - 2)))
            else:
                chapter_names.append(cotc_name.replace("%d", str(chapter_id - 5)))

        return chapter_names

    @staticmethod
    def select_story_chapters(
        save_file: core.SaveFile, chapters: list[int] | None = None
    ) -> list[int] | None:
        chapter_names = StoryChapters.get_chapter_names(save_file, chapters)

        if chapter_names is None:
            return None

        selected_chapters, _ = dialog_creator.ChoiceInput.from_reduced(
            chapter_names, dialog="select_story_chapters"
        ).multiple_choice(localized_options=False)

        return selected_chapters

    @staticmethod
    def get_selected_chapter_progress(max_stages: int = 48) -> int | None:
        progress = dialog_creator.IntInput(
            min=0, max=max_stages
        ).get_input_locale_while("edit_chapter_progress_all", {"max": max_stages})
        if progress is None:
            return None

        return progress

    @staticmethod
    def edit_chapter_progress(
        save_file: core.SaveFile,
        chapter_id: int,
        chapter_name: str,
        clear_amount: int,
        clear_amount_choose: int,
    ) -> bool:
        max_stages = 48
        chapter = save_file.story.get_real_chapters()[chapter_id]
        progress = dialog_creator.IntInput(
            min=0, max=max_stages
        ).get_input_locale_while(
            "edit_chapter_progress",
            {"max": max_stages, "chapter_name": chapter_name},
        )
        if progress is None:
            return False
        clear_amounts = [1] * progress
        if clear_amount_choose == 0:
            clear_amount2 = core.EventChapters.ask_clear_amount()
            if clear_amount2 is None:
                return False
            clear_amounts = [clear_amount2] * progress
        elif clear_amount_choose == 1:
            clear_amounts = [clear_amount] * progress
        elif clear_amount_choose == 2:
            for i in range(progress):
                StoryChapters.print_current_stage(save_file, chapter_id, i)
                clear_amount2 = core.EventChapters.ask_clear_amount()
                if clear_amount2 is None:
                    return False
                clear_amounts[i] = clear_amount2

        chapter.apply_progress(progress, clear_amounts)
        return progress != 0

    @staticmethod
    def convert_stage_id(index: int) -> int:
        if index == 46:
            return 46
        if index == 47:
            return 47
        index = 45 - index
        return index

    @staticmethod
    def ask_clear_count() -> int | None:
        clear_count = dialog_creator.IntInput(
            min=0,
            max=core.core_data.max_value_manager.get("stage_clear_count"),
        ).get_input_locale_while("edit_stage_clear_count", {})

        return clear_count

    @staticmethod
    def ask_if_individual_clear_counts() -> bool | None:
        options = ["individual_clear_counts", "all_clear_counts"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="individual_clear_counts_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return None
        choice -= 1
        return choice == 0

    @staticmethod
    def edit_stage_clear_count(
        save_file: core.SaveFile, chapter_id: int, stage_id: int
    ):
        chapter = save_file.story.get_real_chapters()[chapter_id]
        stage = chapter.stages[stage_id]
        clear_count = StoryChapters.ask_clear_count()
        if clear_count is None:
            return
        stage.clear_times = clear_count

    def clear_previous_chapters(self, chapter_id: int):
        chapters = self.get_real_chapters()
        """
        0: eoc 1
        1: eoc 2 - requires eoc 1
        2: eoc 3 - requires eoc 1 + eoc 2
        3: itf 1 - requires eoc 1
        4: itf 2 - requires eoc 1 + itf 1
        5: itf 3 - requires eoc 1 + itf 1 + itf 2
        6: cotc 1 - requires eoc 1 + itf 1
        7: cotc 2 - requires eoc 1 + itf 1 + cotc 1
        8: cotc 3 - requires eoc 1 + itf 1 + cotc 1 + cotc 2

        """
        if chapter_id == 1:  # eoc 2
            chapters[0].clear_chapter()
        elif chapter_id == 2:  # eoc 3
            chapters[0].clear_chapter()
            chapters[1].clear_chapter()
        elif chapter_id == 3:  # itf 1
            chapters[0].clear_chapter()
        elif chapter_id == 4:  # itf 2
            chapters[0].clear_chapter()
            chapters[3].clear_chapter()
        elif chapter_id == 5:  # itf 3
            chapters[0].clear_chapter()
            chapters[3].clear_chapter()
            chapters[4].clear_chapter()
        elif chapter_id == 6:  # cotc 1
            chapters[0].clear_chapter()
            chapters[3].clear_chapter()
        elif chapter_id == 7:  # cotc 2
            chapters[0].clear_chapter()
            chapters[3].clear_chapter()
            chapters[6].clear_chapter()
        elif chapter_id == 8:  # cotc 3
            chapters[0].clear_chapter()
            chapters[3].clear_chapter()
            chapters[6].clear_chapter()
            chapters[7].clear_chapter()

    @staticmethod
    def print_current_chapter(save_file: core.SaveFile, chapter_id: int):
        chapter_names = StoryChapters.get_chapter_names(save_file)
        if chapter_names is None:
            return
        chapter_name = chapter_names[chapter_id]
        color.ColoredText.localize("current_chapter", chapter_name=chapter_name)

    @staticmethod
    def print_current_treasure_group(
        save_file: core.SaveFile, chapter_id: int, treasure_group_id: int
    ):
        chapter_type = StoryChapters.get_chapter_type_from_index(chapter_id)

        treasure_group_names = TreasureGroupNames(
            save_file, chapter_type
        ).treasure_group_names
        if treasure_group_names is None:
            return
        treasure_group_name = treasure_group_names[treasure_group_id]
        color.ColoredText.localize(
            "current_treasure_group", treasure_group_name=treasure_group_name
        )

    @staticmethod
    def clear_story(save_file: core.SaveFile):
        story = save_file.story
        story.edit_chapters(
            save_file,
        )

    def edit_chapters(self, save_file: core.SaveFile):
        chapters = self.get_real_chapters()
        names = StoryChapters.get_chapter_names(save_file)
        if names is None:
            return

        map_choices = StoryChapters.select_story_chapters(save_file)
        if not map_choices:
            return

        clear_type_choice = dialog_creator.ChoiceInput.from_reduced(
            ["clear_whole_chapters", "clear_specific_stages"],
            dialog="select_clear_type",
            single_choice=True,
        ).single_choice()
        if clear_type_choice is None:
            return
        clear_type_choice -= 1

        modify_clear_amounts = dialog_creator.YesNoInput().get_input_once(
            "modify_clear_amounts"
        )
        if modify_clear_amounts is None:
            return
        clear_amount = 1
        clear_amount_type = -1
        if modify_clear_amounts:
            if len(map_choices) == 1:
                clear_amount_type = 0
            else:
                options = ["clear_amount_chapter", "clear_amount_all"]
                if clear_type_choice == 1:
                    options.append("clear_amount_stages")
                clear_amount_type = dialog_creator.ChoiceInput.from_reduced(
                    options, dialog="select_clear_amount_type", single_choice=True
                ).single_choice()
                if clear_amount_type is None:
                    return
                clear_amount_type -= 1

            if clear_amount_type == 1:
                clear_amount = core.EventChapters.ask_clear_amount()
                if clear_amount is None:
                    return

        for id in map_choices:
            stage_names = StageNames(
                save_file, chapter=str(self.get_chapter_type_from_index(id))
            )
            stage_names = stage_names.stage_names
            if stage_names is None:
                return

            new_stage_names: list[str] = []
            for i in range(48):
                index_stage_id = StoryChapters.convert_stage_id(i)
                new_stage_names.append(stage_names[index_stage_id])
            stage_names = new_stage_names
            map_name = names[id]
            color.ColoredText.localize("current_sol_chapter", name=map_name, id=id)
            if clear_type_choice:
                stages = core.EventChapters.ask_stages_stage_names(stage_names)
                if stages is None:
                    return
            else:
                stages = list(range(48))

            if clear_amount_type == 0:
                clear_amount = core.EventChapters.ask_clear_amount()
                if clear_amount is None:
                    return

            could_unclear_stages = False

            if chapters[id].progress > max(stages) + 1:
                could_unclear_stages = True

            for stage in range(max(stages) + 1, 48):
                if chapters[id].stages[stage].clear_times:
                    could_unclear_stages = True

            if could_unclear_stages:
                unclear_other_stages = dialog_creator.YesNoInput().get_input_once(
                    "unclear_other_stages"
                )
                if unclear_other_stages is None:
                    return
            else:
                unclear_other_stages = False

            if unclear_other_stages:
                chapters[id].progress = 0
                for stage in range(max(stages), 48):
                    chapters[id].stages[stage].clear_times = 0

            for stage in stages:
                if clear_amount_type == 2:
                    stage_name = stage_names[stage]
                    color.ColoredText.localize(
                        "current_sol_stage", name=stage_name, id=stage
                    )
                if clear_amount_type == 2:
                    clear_amount = core.EventChapters.ask_clear_amount()
                    if clear_amount is None:
                        return
                self.clear_stage(
                    id,
                    stage,
                    overwrite_clear_progress=True,
                    clear_amount=clear_amount,
                    chapters=chapters,
                )

        color.ColoredText.localize("map_chapters_edited")

    @staticmethod
    def ask_treasure_level(save_file: core.SaveFile) -> int | None:
        treasure_text = core.core_data.get_treasure_text(save_file).treasure_text
        if treasure_text is None:
            return None
        if len(treasure_text) < 3:
            return None
        options = [
            "no_treasure",
            treasure_text[0],
            treasure_text[1],
            treasure_text[2],
            "custom_treasure_level",
        ]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="treasure_level_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return None
        choice -= 1

        max_treasure_level = core.core_data.max_value_manager.get("treasure_level")

        if choice == 4:
            treasure_level = dialog_creator.IntInput(
                min=0, max=max_treasure_level
            ).get_input_locale_while("custom_treasure_level_dialog", {})
            if treasure_level is None:
                return None
            return treasure_level

        return choice

    @staticmethod
    def get_per_chapter(chapters: list[int]) -> int | None:
        if len(chapters) == 1:
            return 0

        options = ["per_chapter", "all_selected_chapters"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="edit_per_chapter", single_choice=True
        ).single_choice()
        if choice is None:
            return None
        choice -= 1
        return choice

    @staticmethod
    def edit_treasures_whole_chapters(save_file: core.SaveFile, chapters: list[int]):
        choice = StoryChapters.get_per_chapter(chapters)
        if choice is None:
            return

        if choice == 0:
            for chapter_id in chapters:
                StoryChapters.print_current_chapter(save_file, chapter_id)
                chapter = save_file.story.get_real_chapters()[chapter_id]
                treasure_level = StoryChapters.ask_treasure_level(save_file)
                if treasure_level is None:
                    return
                for stage in chapter.get_valid_treasure_stages():
                    stage.set_treasure(treasure_level)
        else:
            treasure_level = StoryChapters.ask_treasure_level(save_file)
            if treasure_level is None:
                return
            for chapter_id in chapters:
                chapter = save_file.story.get_real_chapters()[chapter_id]
                for stage in chapter.get_valid_treasure_stages():
                    stage.set_treasure(treasure_level)

    @staticmethod
    def get_chapter_type_from_index(index: int) -> int:
        if index < 3:
            return 0
        if index < 6:
            return 1
        return 2

    @staticmethod
    def select_stages(save_file: core.SaveFile, chapter_id: int) -> list[int] | None:
        options = ["select_stage_by_id", "select_stage_by_name"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="select_stage_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return None
        choice -= 1

        if choice == 0:
            stage_ids = dialog_creator.RangeInput(48, 1).get_input_locale(
                "select_stage_id", {}
            )
            if stage_ids is None:
                return None
            stage_ids = [stage_id - 1 for stage_id in stage_ids]
            return stage_ids

        chapter_type = StoryChapters.get_chapter_type_from_index(chapter_id)
        stage_names = StageNames(save_file, str(chapter_type)).stage_names
        if not stage_names:
            return None
        new_stage_names: list[str] = []
        for i in range(48):
            index_stage_id = StoryChapters.convert_stage_id(i)
            new_stage_names.append(stage_names[index_stage_id])
        selected_stages, _ = dialog_creator.ChoiceInput.from_reduced(
            new_stage_names, dialog="select_stages_name"
        ).multiple_choice(localized_options=False)

        if not selected_stages:
            return None

        return selected_stages

    @staticmethod
    def edit_treasures_individual_stages(save_file: core.SaveFile, chapters: list[int]):
        choice = StoryChapters.get_per_chapter(chapters)
        if choice is None:
            return
        if choice == 0:
            for chapter_id in chapters:
                StoryChapters.print_current_chapter(save_file, chapter_id)
                chapter = save_file.story.get_real_chapters()[chapter_id]
                stage_ids = StoryChapters.select_stages(save_file, chapter_id)
                if stage_ids is None:
                    return
                treasure_level = StoryChapters.ask_treasure_level(save_file)
                if treasure_level is None:
                    return
                for stage_id in stage_ids:
                    real_stage_id = StoryChapters.convert_stage_id(stage_id)
                    chapter.set_treasure(real_stage_id, treasure_level)
        else:
            stage_ids = StoryChapters.select_stages(save_file, 0)
            if stage_ids is None:
                return
            treasure_level = StoryChapters.ask_treasure_level(save_file)
            if treasure_level is None:
                return
            for chapter_id in chapters:
                chapter = save_file.story.get_real_chapters()[chapter_id]
                for stage_id in stage_ids:
                    real_stage_id = StoryChapters.convert_stage_id(stage_id)
                    chapter.set_treasure(real_stage_id, treasure_level)

    @staticmethod
    def edit_treasures_groups(save_file: core.SaveFile, chapters: list[int]):
        for chapter_id in chapters:
            StoryChapters.print_current_chapter(save_file, chapter_id)
            chapter = save_file.story.get_real_chapters()[chapter_id]
            chapter_type = StoryChapters.get_chapter_type_from_index(chapter_id)
            treasure_group_data = TreasureGroupData(
                save_file, chapter_type
            ).treasure_group_data
            treasure_group_names = TreasureGroupNames(
                save_file, chapter_type
            ).treasure_group_names
            if not treasure_group_data or not treasure_group_names:
                return
            treasure_group_names_new: list[str] = []
            for i in range(len(treasure_group_data)):
                treasure_group_names_new.append(treasure_group_names[i])

            selected_treasure_groups, _ = dialog_creator.ChoiceInput.from_reduced(
                treasure_group_names_new, dialog="select_treasure_groups"
            ).multiple_choice(localized_options=False)

            if not selected_treasure_groups:
                return

            options = ["group_individual", "group_all_at_once"]
            choice = dialog_creator.ChoiceInput.from_reduced(
                options, dialog="select_treasure_groups_individual"
            ).single_choice()
            if choice is None:
                return
            choice -= 1

            if choice == 0:
                for treasure_group_id in selected_treasure_groups:
                    StoryChapters.print_current_treasure_group(
                        save_file, chapter_id, treasure_group_id
                    )
                    treasure_level = StoryChapters.ask_treasure_level(save_file)
                    if treasure_level is None:
                        return
                    treasure_group = treasure_group_data[treasure_group_id]
                    for stage_id in treasure_group:
                        chapter.set_treasure(stage_id, treasure_level)

            else:
                treasure_level = StoryChapters.ask_treasure_level(save_file)
                if treasure_level is None:
                    return

                for treasure_group_id in selected_treasure_groups:
                    treasure_group = treasure_group_data[treasure_group_id]
                    for stage_id in treasure_group:
                        chapter.set_treasure(stage_id, treasure_level)

    @staticmethod
    def edit_treasures(save_file: core.SaveFile):
        selected_chapters = StoryChapters.select_story_chapters(save_file)
        if not selected_chapters:
            return
        options = ["whole_chapters", "individual_stages", "treasure_groups"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="treasure_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return
        choice -= 1

        if choice == 0:
            StoryChapters.edit_treasures_whole_chapters(save_file, selected_chapters)
        elif choice == 1:
            StoryChapters.edit_treasures_individual_stages(save_file, selected_chapters)
        elif choice == 2:
            StoryChapters.edit_treasures_groups(save_file, selected_chapters)

        color.ColoredText.localize("treasures_edited")

    @staticmethod
    def edit_itf_timed_scores(save_file: core.SaveFile):
        selected_chapters = StoryChapters.select_story_chapters(
            save_file, chapters=[3, 4, 5]
        )
        if not selected_chapters:
            return
        options = ["whole_chapters", "individual_stages"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="itf_timed_scores_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return
        choice -= 1

        selected_chapters = [chapter_id + 3 for chapter_id in selected_chapters]

        if choice == 0:
            StoryChapters.edit_itf_timed_scores_whole_chapters(
                save_file, selected_chapters
            )
        elif choice == 1:
            StoryChapters.edit_itf_timed_scores_individual_stages(
                save_file, selected_chapters
            )

        color.ColoredText.localize("itf_timed_scores_edited")

    @staticmethod
    def edit_itf_timed_scores_whole_chapters(
        save_file: core.SaveFile, chapters: list[int]
    ):
        choice = StoryChapters.get_per_chapter(chapters)
        if choice is None:
            return

        if choice == 0:
            for chapter_id in chapters:
                print(chapter_id)
                StoryChapters.print_current_chapter(save_file, chapter_id)
                chapter = save_file.story.get_real_chapters()[chapter_id]
                score = dialog_creator.IntInput(
                    min=0,
                    max=core.core_data.max_value_manager.get("itf_timed_score"),
                ).get_input_locale_while("itf_timed_score_dialog", {})
                if score is None:
                    return
                for stage in chapter.get_valid_treasure_stages():
                    stage.itf_timed_score = score
        else:
            score = dialog_creator.IntInput(
                min=0,
                max=core.core_data.max_value_manager.get("itf_timed_score"),
            ).get_input_locale_while("itf_timed_score_dialog", {})
            if score is None:
                return
            for chapter_id in chapters:
                chapter = save_file.story.get_real_chapters()[chapter_id]
                for stage in chapter.get_valid_treasure_stages():
                    stage.itf_timed_score = score

    @staticmethod
    def print_current_stage(save_file: core.SaveFile, chapter_id: int, stage_id: int):
        chapter_names = StoryChapters.get_chapter_names(save_file)
        if chapter_names is None:
            return
        chapter_name = chapter_names[chapter_id]
        chapter_type = StoryChapters.get_chapter_type_from_index(chapter_id)
        stage_names = StageNames(save_file, str(chapter_type)).stage_names
        if stage_names is None:
            return
        stage_id = StoryChapters.convert_stage_id(stage_id)
        stage_name = stage_names[stage_id]
        color.ColoredText.localize(
            "current_stage", chapter_name=chapter_name, stage_name=stage_name
        )

    @staticmethod
    def edit_itf_timed_scores_individual_stages(
        save_file: core.SaveFile, chapters: list[int]
    ):
        choice = StoryChapters.get_per_chapter(chapters)
        if choice is None:
            return
        options = ["individual_stages", "all_selected_stages"]
        choice2 = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="itf_timed_scores_individual_dialog",
            single_choice=True,
        ).single_choice()
        if choice2 is None:
            return
        choice2 -= 1

        if choice == 0:
            for chapter_id in chapters:
                StoryChapters.print_current_chapter(save_file, chapter_id)
                chapter = save_file.story.get_real_chapters()[chapter_id]
                stage_ids = StoryChapters.select_stages(save_file, chapter_id)
                if stage_ids is None:
                    return
                if choice2 == 0:
                    for stage_id in stage_ids:
                        StoryChapters.print_current_stage(
                            save_file, chapter_id, stage_id
                        )

                        score = dialog_creator.IntInput(
                            min=0,
                            max=core.core_data.max_value_manager.get("itf_timed_score"),
                        ).get_input_locale_while("itf_timed_score_dialog", {})
                        if score is None:
                            return
                        chapter.stages[stage_id].itf_timed_score = score
                elif choice2 == 1:
                    score = dialog_creator.IntInput(
                        min=0,
                        max=core.core_data.max_value_manager.get("itf_timed_score"),
                    ).get_input_locale_while("itf_timed_score_dialog", {})
                    if score is None:
                        return
                    for stage_id in stage_ids:
                        chapter.stages[stage_id].itf_timed_score = score
        else:
            stage_ids = StoryChapters.select_stages(save_file, 3)
            if stage_ids is None:
                return
            if choice2 == 0:
                for stage_id in stage_ids:
                    StoryChapters.print_current_stage(save_file, 3, stage_id)
                    score = dialog_creator.IntInput(
                        min=0,
                        max=core.core_data.max_value_manager.get("itf_timed_score"),
                    ).get_input_locale_while("itf_timed_score_dialog", {})
                    if score is None:
                        return
                    for chapter_id in chapters:
                        chapter = save_file.story.get_real_chapters()[chapter_id]
                        chapter.stages[stage_id].itf_timed_score = score
            elif choice2 == 1:
                score = dialog_creator.IntInput(
                    min=0,
                    max=core.core_data.max_value_manager.get("itf_timed_score"),
                ).get_input_locale_while("itf_timed_score_dialog", {})
                if score is None:
                    return
                for chapter_id in chapters:
                    chapter = save_file.story.get_real_chapters()[chapter_id]
                    for stage_id in stage_ids:
                        chapter.stages[stage_id].itf_timed_score = score


class StageNames:
    def __init__(self, save_file: core.SaveFile, chapter: str, max_stages: int = 48):
        self.save_file = save_file
        self.chapter = chapter
        self.max_stages = max_stages
        self.stage_names = self.get_stage_names()

    def get_file_name(self) -> str:
        if self.chapter.isdigit():
            return (
                f"StageName{self.chapter}_{core.core_data.get_lang(self.save_file)}.csv"
            )
        return f"StageName_{self.chapter}_{core.core_data.get_lang(self.save_file)}.csv"

    def get_stage_names(self) -> list[str] | None:
        file_name = self.get_file_name()
        gdg = core.core_data.get_game_data_getter(self.save_file)
        file = gdg.download("resLocal", file_name)
        if file is None:
            return None
        csv = core.CSV(
            file,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        stage_names: list[str] = []
        if self.chapter.isdigit():
            for row in csv:
                stage_names.append(row[0].to_str())
        else:
            for row in csv:
                for value in row:
                    stage_names.append(value.to_str())
        return stage_names[: self.max_stages]

    def get_stage_name(self, stage_id: int) -> str | None:
        if self.stage_names is None:
            return None
        return self.stage_names[stage_id]


class TreasureText:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.treasure_text = self.get_treasure_text()

    def get_tt_file_name(self) -> str:
        return f"Treasure2_{core.core_data.get_lang(self.save_file)}.csv"

    def get_treasure_text(self) -> list[str] | None:
        file_name = self.get_tt_file_name()
        gdg = core.core_data.get_game_data_getter(self.save_file)
        file = gdg.download("resLocal", file_name)
        if file is None:
            return None
        csv = core.CSV(
            file,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        treasure_text: list[str] = []
        for row in csv:
            treasure_text.append(row[0].to_str())
        return treasure_text


class TreasureGroupData:
    def __init__(self, save_file: core.SaveFile, chapter_type: int):
        self.save_file = save_file
        self.chapter_type = chapter_type
        self.treasure_group_data = self.get_treasure_group_data()

    def get_tgd_file_name(self) -> str:
        if self.chapter_type == 0:
            return "treasureData0.csv"
        if self.chapter_type == 1:
            return "treasureData1.csv"
        if self.chapter_type == 2:
            return "treasureData2_0.csv"
        return ""

    def get_treasure_group_data(self) -> list[list[int]] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        file = gdg.download("DataLocal", self.get_tgd_file_name())
        if file is None:
            return None
        csv = core.CSV(file)
        treasure_group_data: list[list[int]] = []
        for row in csv.lines[11:22]:
            treasure_group_data.append(
                [value.to_int() for value in row if value.to_int() != -1]
            )

        return treasure_group_data


class TreasureGroupNames:
    def __init__(self, save_file: core.SaveFile, chapter_type: int):
        self.save_file = save_file
        self.chapter_type = chapter_type
        self.treasure_group_names = self.get_treasure_group_names()

    def get_tgn_file_name(self) -> str:
        lang = core.core_data.get_lang(self.save_file)
        if self.chapter_type == 0:
            return f"Treasure3_0_{lang}.csv"
        if self.chapter_type == 1:
            return f"Treasure3_1_{lang}.csv"
        if self.chapter_type == 2:
            return f"Treasure3_2_0_{lang}.csv"
        return ""

    def get_treasure_group_names(self) -> list[str] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        file = gdg.download("resLocal", self.get_tgn_file_name())
        if file is None:
            return None
        csv = core.CSV(
            file,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        treasure_group_names: list[str] = []
        for row in csv:
            treasure_group_names.append(row[0].to_str())
        return treasure_group_names



from __future__ import annotations
from io import BytesIO
from typing import Any, Callable

from bcsfe.cli import color, dialog_creator

import tarfile

from bcsfe import core


class GameDataGetter:
    def __init__(
        self, cc: core.CountryCode, gv: core.GameVersion, do_print: bool = True
    ):
        self.repo_url = core.core_data.config.get_game_data_repo()
        self.print = do_print
        self.lang = core.core_data.config.get_str(core.ConfigKey.LOCALE)
        self.cc = cc.get_cc_lang()
        self.real_cc = cc
        self.gv = gv
        self.cc = self.cc if not self.cc.is_lang() else self.real_cc
        self.version, exact_match = self.find_gv(self.cc, gv)

        self.all_versions = None
        self.url = None
        self.filepath = None

        if exact_match:
            return

        self.metadata = self.get_metadata()
        if self.metadata is None:
            return
        self.all_versions = self.get_versions(self.metadata)
        self.url = self.metadata.get("base_url")
        if self.all_versions is not None:
            self.version, self.filepath = self.get_version(self.all_versions, self.cc)

    def find_gv(
        self, cc: core.CountryCode, gv: core.GameVersion
    ) -> tuple[str | None, bool]:
        versions = GameDataGetter.get_all_downloaded_versions().get(cc.get_code())
        if versions is None:
            return None, False

        versions_int = [
            core.GameVersion.from_string(version).game_version for version in versions
        ]

        versions_int.sort()

        for version in versions_int:
            if version >= gv.game_version:
                return core.GameVersion(version).to_string(), version == gv.game_version
        return None, False

    def does_save_version_match(self, save_file: core.SaveFile) -> bool:
        if self.version is None:
            return False

        return save_file.game_version == self.version

    def get_version(
        self, versions: dict[str, dict[str, str]], cc: core.CountryCode
    ) -> tuple[str | None, str | None]:
        cc_versions = versions.get(cc.get_code())
        if cc_versions is None:
            return None, None
        if not cc_versions:
            return None, None
        gv_string = self.gv.to_string()
        if gv_string not in cc_versions:
            cc_version_keys = list(cc_versions.keys())
            cc_version_keys.sort()
            for version in cc_version_keys:
                if (
                    core.GameVersion.from_string(version).game_version
                    >= self.gv.game_version
                ):
                    return version, cc_versions[version]
            return cc_version_keys[-1], cc_versions[cc_version_keys[-1]]
        return gv_string, cc_versions[gv_string]

    def get_metadata(self, show_alt: bool = True) -> dict[str, Any] | None:
        response = core.RequestHandler(self.repo_url).get()
        if response is None:
            if (
                self.repo_url
                == core.core_data.config.get_default(core.ConfigKey.GAME_DATA_REPO)
                and show_alt
            ):
                alt = "https://gitlab.com/fieryhenry/bcdata/-/raw/main/metadata.json"
                res = dialog_creator.YesNoInput().get_input_once(
                    "use_alternative_repo",
                    {
                        "repo": "https://gitlab.com/fieryhenry/bcdata/-/raw/main/metadata.json"
                    },
                )
                if res:
                    core.core_data.config.set(core.ConfigKey.GAME_DATA_REPO, alt)
                    self.repo_url = alt
                    return self.get_metadata(show_alt=False)

            return None
        try:
            data = response.json()
        except core.JSONDecodeError as e:
            print(e)
            return None
        return data

    def get_versions(self, metdata: dict[str, Any]) -> dict[str, dict[str, str]] | None:
        return metdata.get("versions")

    def get_packname(self, packname: str) -> str:
        if packname != "resLocal":
            return packname
        if self.cc != core.CountryCodeType.EN:
            return packname
        langs = core.CountryCode.get_langs()
        if self.lang in langs:
            return f"{packname}_{self.lang}"
        return packname

    @staticmethod
    def get_game_data_dir() -> core.Path:
        return core.Path.get_documents_folder().add("game_data")

    def get_file_path(self, pack_name: str, file_name: str) -> core.Path | None:
        pack_name = self.get_packname(pack_name)
        path = self.get_version_path()
        if path is None:
            return None
        return path.add(pack_name).generate_dirs().add(file_name)

    def download_version_data(self):
        if self.url is None or self.filepath is None or self.version is None:
            return None
        url = self.url + self.filepath

        if self.print:
            color.ColoredText.localize("downloading_compressed_data", url=url)

        downloaded_data = core.RequestHandler(url).get()
        if downloaded_data is None:
            if self.print:
                color.ColoredText.localize("no_internet")
            return None

        archive = tarfile.open(
            name=self.filepath, fileobj=BytesIO(downloaded_data.content)
        )

        outdir = (
            GameDataGetter.get_game_data_dir().add(self.cc.get_code()).add(self.version)
        ).generate_dirs()

        archive.extractall(outdir.path)

        outdir.add("downloaded").write(core.Data())

        return True

    def get_version_path(self) -> core.Path | None:
        if self.version is None:
            return None
        return (
            GameDataGetter.get_game_data_dir().add(self.cc.get_code()).add(self.version)
        ).generate_dirs()

    def has_downloaded(self) -> bool:
        path = self.get_version_path()
        if path is None:
            return False
        return path.add("downloaded").exists()

    def get_file(self, pack_name: str, file_name: str) -> core.Data | bool:
        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return False

        if path.exists():
            return path.read()
        else:
            if self.has_downloaded():
                return True
            if self.download_version_data() is None:
                return False

            path = self.get_file_path(pack_name, file_name)
            if path is None:
                return False

            if path.exists():
                return path.read()
            return self.has_downloaded()

    def save_file(self, pack_name: str, file_name: str) -> core.Data | bool:
        pack_name = self.get_packname(pack_name)
        data = self.get_file(pack_name, file_name)
        if isinstance(data, bool):
            return data

        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return False
        data.to_file(path)
        return data

    def save_file_data(
        self, pack_name: str, file_name: str, data: core.Data
    ) -> core.Data | None:
        pack_name = self.get_packname(pack_name)

        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return None
        data.to_file(path)
        return data

    def is_downloaded(self, pack_name: str, file_name: str) -> bool:
        pack_name = self.get_packname(pack_name)
        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return False
        return path.exists()

    def download_from_path(
        self, path: str, retries: int = 2, display_text: bool = True
    ) -> core.Data | None:
        pack_name, file_name = path.split("/")
        pack_name = self.get_packname(pack_name)
        return self.download(pack_name, file_name, retries, display_text)

    def download(
        self,
        pack_name: str,
        file_name: str,
        retries: int = 2,
        display_text: bool = True,
    ) -> core.Data | None:
        retries -= 1
        pack_name = self.get_packname(pack_name)

        if self.is_downloaded(pack_name, file_name):
            path = self.get_file_path(pack_name, file_name)
            if path is None:
                return None
            try:
                return path.read()
            except FileNotFoundError:
                return None

        if retries == 0:
            return None

        version = self.version

        if version is None:
            if display_text:
                self.print_no_file(pack_name, file_name)
            return None

        if display_text and not self.has_downloaded():
            color.ColoredText.localize(
                "downloading",
                file_name=file_name,
                pack_name=pack_name,
                country_code=self.cc.get_code(),
                version=version,
            )
        data = self.save_file(pack_name, file_name)
        if isinstance(data, bool):
            if not data and display_text:
                self.print_no_file(pack_name, file_name)
            return None

        data = self.download(pack_name, file_name, retries, display_text)
        if data is None:
            if display_text:
                self.print_no_file(pack_name, file_name)
            return None
        return data

    def download_all(
        self,
        pack_name: str,
        file_names: list[str],
        display_text: bool = True,
    ) -> list[tuple[str, core.Data] | None]:
        pack_name = self.get_packname(pack_name)

        callables: list[Callable[..., Any]] = []
        args: list[tuple[str, str, int, bool]] = []
        for file_name in file_names:
            callables.append(self.download)
            args.append((pack_name, file_name, 2, display_text))
        core.thread_run_many(callables, args)
        data_list: list[tuple[str, core.Data] | None] = []
        for file_name in file_names:
            path = self.get_file_path(pack_name, file_name)
            if path is None:
                data_list.append(None)
            elif not path.exists():
                data_list.append(None)
            else:
                data_list.append((file_name, path.read()))
        return data_list

    @staticmethod
    def get_all_downloaded_versions() -> dict[str, list[str]]:
        versions: dict[str, list[str]] = {}
        for cc in core.CountryCode.get_all_str():
            dir = GameDataGetter.get_game_data_dir().add(cc)
            if not dir.exists():
                continue
            for version in GameDataGetter.get_game_data_dir().add(cc).get_dirs():
                if not version.exists():
                    continue
                if not version.add("downloaded").exists():
                    continue
                if cc in versions:
                    versions[cc].append(version.basename())
                else:
                    versions[cc] = [version.basename()]

        return versions

    @staticmethod
    def delete_old_versions(to_keep: int) -> None:
        versions = GameDataGetter.get_all_downloaded_versions()
        for cc, cc_versions in versions.items():
            cc_versions.sort(reverse=True)
            to_keep = min(to_keep, len(cc_versions))
            for version in cc_versions[to_keep:]:
                path = GameDataGetter.get_game_data_dir().add(cc).add(version)
                path.remove()

    def print_no_file(self, packname: str, file_name: str) -> None:
        if self.version is None:
            color.ColoredText.localize("failed_to_get_game_versions")
        else:
            color.ColoredText.localize(
                "failed_to_download_game_data",
                file_name=file_name,
                pack_name=packname,
                country_code=self.cc.get_code(),
                version=self.version,
                url=self.url,
            )



from __future__ import annotations
from bcsfe import core
from typing import Any

from bcsfe.cli import color


class GamblingEvent:
    def __init__(
        self,
        completed: dict[int, bool],
        values: dict[int, dict[int, int]],
        start_times: dict[int, int | float],
    ):
        self.completed = completed
        self.values = values
        self.start_times = start_times

    @staticmethod
    def init() -> GamblingEvent:
        return GamblingEvent({}, {}, {})

    @staticmethod
    def read(data: core.Data, game_version: core.GameVersion) -> GamblingEvent:
        total = data.read_short()
        completed: dict[int, bool] = {}

        for _ in range(total):
            key = data.read_short()
            completed[key] = data.read_bool()

        total = data.read_short()
        values: dict[int, dict[int, int]] = {}

        for _ in range(total):
            key = data.read_short()
            if key not in values:
                values[key] = {}

            total2 = data.read_short()
            for _ in range(total2):
                key2 = data.read_short()

                values[key][key2] = data.read_short()

        total = data.read_short()
        start_times: dict[int, int | float] = {}

        for _ in range(total):
            key = data.read_short()

            if game_version < 90100:
                value = data.read_double()
            else:
                value = data.read_int()

            start_times[key] = value

        return GamblingEvent(completed, values, start_times)

    def write(self, data: core.Data, game_version: core.GameVersion):
        data.write_short(len(self.completed))
        data.write_short_bool_dict(self.completed, write_length=False)

        data.write_short(len(self.values))

        for key, value in self.values.items():
            data.write_short(key)
            data.write_short(len(value))

            for key2, value2 in value.items():
                data.write_short(key2)
                data.write_short(value2)

        data.write_short(len(self.start_times))
        for key, value in self.start_times.items():
            data.write_short(key)

            # this is a bad conversion, since float is timestamp i assume and int as the date as YYYMMDD. FIXME
            if game_version < 90100:
                data.write_double(float(value))
            else:
                data.write_int(int(value))

    def serialize(self) -> dict[str, Any]:
        return {
            "completed": self.completed,
            "values": self.values,
            "start_times": self.start_times,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> GamblingEvent:
        return GamblingEvent(
            data.get("completed", {}),
            data.get("values", {}),
            data.get("start_times", {}),
        )

    def reset(self):
        self.completed = {}
        self.values = {}
        # TODO: check start times
        self.start_times = {}

    @staticmethod
    def reset_events(save_file: core.SaveFile):
        save_file.wildcat_slots.reset()
        color.ColoredText.localize("reset_wildcat_slots")
        save_file.cat_scratcher.reset()
        color.ColoredText.localize("reset_cat_scratcher")



from __future__ import annotations
from bcsfe import core


class Fruit:
    def __init__(
        self,
        id: int,
        seed: bool,
        group: int,
        sort: int,
        require: int | None = None,
        text: str | None = None,
        grow_up: list[int] | None = None,
    ):
        self.id = id
        self.seed = seed
        self.group = group
        self.sort = sort
        self.require = require
        self.text = text
        self.grow_up = grow_up


class Matatabi:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.matatabi = self.__get_matatabi()
        self.gatya_item_names = core.core_data.get_gatya_item_names(
            self.save_file
        )

    def __get_matatabi(self) -> list[Fruit] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "Matatabi.tsv")
        if data is None:
            return None
        csv = core.CSV(data, "\t")
        matatabi: list[Fruit] = []
        for line in csv.lines[1:]:
            id = line[0].to_int()
            seed = line[1].to_bool()
            group = line[2].to_int()
            sort = line[3].to_int()
            if len(line) > 4:
                require = line[4].to_int()
            else:
                require = None
            if len(line) > 5:
                text = line[5].to_str()
            else:
                text = None
            if len(line) > 6:
                grow_up = [item.to_int() for item in line[6:]]
            else:
                grow_up = None
            matatabi.append(
                Fruit(id, seed, group, sort, require, text, grow_up)
            )

        return matatabi

    def get_names(self) -> list[str | None] | None:
        if self.matatabi is None:
            return None

        ids = [fruit.id for fruit in self.matatabi]
        names = [self.gatya_item_names.get_name(id) for id in ids]
        return names



from bcsfe.core.io import (
    bc_csv,
    path,
    data,
    command,
    yaml,
    config,
    json_file,
    save,
    thread_helper,
    root_handler,
    adb_handler,
    git_handler,
    waydroid,
)

__all__ = [
    "bc_csv",
    "path",
    "data",
    "command",
    "yaml",
    "config",
    "json_file",
    "save",
    "thread_helper",
    "root_handler",
    "adb_handler",
    "git_handler",
    "waydroid",
]



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import dialog_creator


class ChallengeChapters:
    def __init__(self, chapters: core.Chapters):
        self.chapters = chapters
        self.scores: list[int] = []
        self.shown_popup: bool = False

    @staticmethod
    def init() -> ChallengeChapters:
        return ChallengeChapters(core.Chapters.init())

    @staticmethod
    def read(data: core.Data) -> ChallengeChapters:
        ch = core.Chapters.read(data)
        return ChallengeChapters(ch)

    def write(self, data: core.Data):
        self.chapters.write(data)

    def read_scores(self, data: core.Data):
        total_scores = data.read_int()
        self.scores = [data.read_int() for _ in range(total_scores)]

    def write_scores(self, data: core.Data):
        data.write_int(len(self.scores))
        for score in self.scores:
            data.write_int(score)

    def read_popup(self, data: core.Data):
        self.shown_popup = data.read_bool()

    def write_popup(self, data: core.Data):
        data.write_bool(self.shown_popup)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": self.chapters.serialize(),
            "scores": self.scores,
            "shown_popup": self.shown_popup,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ChallengeChapters:
        challenge = ChallengeChapters(
            core.Chapters.deserialize(data.get("chapters", {})),
        )
        challenge.scores = data.get("scores", [])
        challenge.shown_popup = data.get("shown_popup", False)
        return challenge

    def __repr__(self):
        return f"Challenge({self.chapters})"

    def __str__(self):
        return self.__repr__()

    def edit_score(self):
        if not self.scores:
            self.scores = [0]
        self.scores[0] = dialog_creator.SingleEditor(
            "challenge_score", self.scores[0], None, localized_item=True
        ).edit()
        self.shown_popup = True
        self.chapters.clear_stage(0, 0, 0, False)


def edit_challenge_score(save_file: core.SaveFile):
    save_file.challenge.edit_score()
